package de.paluno.serviceplatform;

import java.util.Vector;

/**
 * 
 * @author Eric Schmieders (eric.schmieders@sse.uni-due.de)
 *
 */
public class DetectingComponent {
	private static final String OUTPUT = "output/";
	private PathCache pathCache;

	private boolean VERBOSE = MonitoringComponent.VERBOSE;
	
	private int e1InstanceExecuted = 0;
	private int e1PossibleServiceInvocations = 0;
	private int e1ServiceInvoked = 0;
	private int e1InstanceWithoutViolation = 0;
	private int e1InstanceWithViolation = 0;
	private int e1AdaptationNotPossible = 0;
	private int e1InstanceWithAdaptation = 0;
	private int e1FalsePositive = 0;
	
	private int e2InstanceExecuted = 0;
	private int e2InstanceWithoutViolation = 0;
	private int e2InstanceWithViolation = 0;
	private int e2AdaptationNotPossible = 0;
	private int e2InstanceWithAdaptation = 0;
	private int e2FalsePositive = 0;

	private int e3InstanceExecuted = 0;
	private int e3PossibleServiceInvocations = 0;
	private int e3ServiceInvoked = 0;
	private int e3InstanceWithoutViolation = 0;
	private int e3InstanceWithViolation = 0;
	private int e3AdaptationNotPossible = 0;
	private int e3InstanceWithAdaptation = 0;
	private int e3FalsePositive = 0;	
	
	private static long RequirementMaxResponseTime = 20000;

	public DetectingComponent(PathCache pathCache) {
		this.pathCache = pathCache;
	}

	public PathCache getPathCache() {
		return pathCache;
	}

	public static long getSumMonitoredRuntime(Vector<Activity> v, int execution) {
		long sum = 0;

		for (Activity a : v)
			sum += a.getMonitoredResponseTime(execution);

		return sum;
	}

	private void runTestE1(int execution, Vector<Activity> p, long e2eResponseTime) {
		print("*** Executing Service Monitoring Test for no." + execution + " ***");
		
		e1InstanceExecuted++;
		e1PossibleServiceInvocations += p.size();
		
		int currentExecutionTime = 0;
		boolean withoutViolation = true;
		
		for (Activity a: p) {
			currentExecutionTime += a.getMonitoredResponseTime(execution);
			e1ServiceInvoked++;
			
			if (a.getMonitoredResponseTime(execution) > a.getAssumption()) {
				print("Violation@'"+a.getName()+"': monitoredResponseTime("+a.getMonitoredResponseTime(execution)+") |/= assumedResponseTime("+a.getAssumption()+")");
				e1InstanceWithViolation++;
				withoutViolation = false;
				if (currentExecutionTime > getRTRequirement()) {
					print("Adaptation Not possible: currentExecutionTime("+currentExecutionTime+") > requirement("+getRTRequirement()+")");
					e1AdaptationNotPossible++;
				} else {
					print("Adaptation possible: currentExecutionTime("+currentExecutionTime+") < requirement("+getRTRequirement()+")");
					e1InstanceWithAdaptation++;
					if (e2eResponseTime <= getRTRequirement()) {
						print("False Positive: monitoredExecutionTime("+e2eResponseTime+") |= requirement("+getRTRequirement()+")");
						e1FalsePositive++;
					}
				}
				break;
			} else
				print("Compliant Execution '"+a.getName()+"': monitoredResponseTime("+a.getMonitoredResponseTime(execution)+") |= assumedResponseTime("+a.getAssumption()+")");
		}
		
		if (withoutViolation) {
			print("Instance Report: Instance "+execution+" executed without any deviations.");
			e1InstanceWithoutViolation++;
		} else 
			print("Instance Report: During execution of instance "+execution+" deviations occur.");
	}
	
	private void runTestE3(int execution, Vector<Activity> p,
			long e2eResponseTime) {
		
		print("*** Executing SPADE Test for no." + execution + " ***");
		
		e3InstanceExecuted ++;
		e3PossibleServiceInvocations += p.size();

		int currentExecutionTime = 0;
		boolean withoutViolation = true;
		
		for (Activity a: p) {
			currentExecutionTime += a.getMonitoredResponseTime(execution);
			e3ServiceInvoked++;
			
			if (a.getMonitoredResponseTime(execution) > a.getAssumption()) {
				print("Violation@'"+a.getName()+"': monitoredResponseTime("+a.getMonitoredResponseTime(execution)+") |/= assumedResponseTime("+a.getAssumption()+")");
				withoutViolation = false;
				if (currentExecutionTime > getRTRequirement()) {
					print("Adaptation Not possible: currentExecutionTime("+currentExecutionTime+") > requirement("+getRTRequirement()+")");
					e3AdaptationNotPossible++;
				} else {
					print("Adaptation possible: currentExecutionTime("+currentExecutionTime+") < requirement("+getRTRequirement()+")");
					
					// Check, ob adaptation
					DetectionResult d = checkRequirements(p, a, execution);
					
					if (d != null) {
						if (e2eResponseTime <= getRTRequirement()) {
							print("False Positive: monitoredExecutionTime("+e2eResponseTime+") |= requirement("+getRTRequirement()+")");
							e3FalsePositive++;
						}
						e3InstanceWithAdaptation++;
						break;
					}
				}
			} else
				print("Compliant Execution '"+a.getName()+"': monitoredResponseTime("+a.getMonitoredResponseTime(execution)+") |= assumedResponseTime("+a.getAssumption()+")");
		}
		
		if (withoutViolation) {
			print("Instance Report: Instance "+execution+" executed without any deviations.");
			e3InstanceWithoutViolation++;
		} else { 
			print("Instance Report: During execution of instance "+execution+" deviations occur.");
			e3InstanceWithViolation ++;
		}
	}
	
	public void runInstance(int execution) {
		print("*** Executing instance no." + execution + " ***");
		
		Vector<Activity> executionPath = pathCache.getExecutionPath();
		
		runTestE1(execution, executionPath, getSumMonitoredRuntime(executionPath,execution));
		runTestE3(execution, executionPath, getSumMonitoredRuntime(executionPath,execution));
	}


	private void print(int i) {
		print(""+i);
	}
	
	private void print(String string) {
		if (VERBOSE)
			System.out.println(string);
	}
	
	private boolean assumptionViolation(Activity current, int execution) {
		if (current.getMonitoredResponseTime(execution) <= current
				.getAssumption()) {
			print("* " + current.getPrint(execution, false, true)
					+ " |= " + current.getPrint(execution, true, false) + "*");
			return false;
		} else {
			print("* " + current.getPrint(execution, false, true)
					+ " |/= " + current.getPrint(execution, true, false) + "*");
			return true;
		}
	}

	static public String printPath(Vector<Activity> v, int execution) {
		String result = "";
		for (Activity a : v)
			result += " " + a.getPrint(execution, true, true);
		return result;
	}

	private DetectionResult checkRequirements(Vector<Activity> executionPath,
			Activity current, int execution) {
		// Pfad bis zur (inkl.) derzeitigen Acitivity ermitteln
		Vector<Activity> pathSoFar = new Vector<Activity>();
		for (Activity a : executionPath) {
			pathSoFar.add(a);
			if (a == current)
				break;
		}

		// Die relevanten Pfade aus dem Cache ermitteln
		Vector<Vector<Activity>> relevantPaths = pathCache
				.getRelevantPaths(pathSoFar);

		print("Check requirement violation for relevant paths:");
		// Prfen ob die relevanten Pfade die Requirements verletzen
		for (Vector<Activity> v : relevantPaths) {
			print("Checking path:" + printPath(v, execution));
			DetectionResult d = checkRequirementForCachedPath(v, current,
					execution);
			if (d.getAdaptationTrigger()) {
				print("CompletedWithAssumptions("
						+ d.getResponseTimeCompletedWithAssumptions()
						+ ") |/= Requirement(" + RequirementMaxResponseTime
						+ ")");
				print("==> TRY TRIGGER ADAPTATION");
				return d;
			}
		}
		return null;
	}

	private DetectionResult checkRequirementForCachedPath(Vector<Activity> v,
			Activity current, int execution) {
		long responseTimeCompletedWithAssumption = 0;
		long responseTimeMonitoringData = 0;
		long responseTimeSoFar = 0;

		boolean useMonitoringData = true;

		for (Activity a : v) {
			if (useMonitoringData) {
				responseTimeSoFar += a.getMonitoredResponseTime(execution);
				responseTimeCompletedWithAssumption += a
						.getMonitoredResponseTime(execution);
			} else {
				responseTimeCompletedWithAssumption += a.getAssumption();
			}

			if (a.sameAs(current))
				useMonitoringData = false;

			responseTimeMonitoringData += a.getMonitoredResponseTime(execution);
		}

		boolean adaptationTrigger = false;
		if (responseTimeCompletedWithAssumption > RequirementMaxResponseTime)
			adaptationTrigger = true;

		return new DetectionResult(responseTimeMonitoringData,
				responseTimeCompletedWithAssumption, responseTimeSoFar,
				adaptationTrigger);
	}

	
	public void setRTRequirement(long executionPathAssumedRT) {
		this.RequirementMaxResponseTime = executionPathAssumedRT;
	}
	
	public long getRTRequirement() {
		return RequirementMaxResponseTime;
	}
	
	
	public int getE2InstanceExecuted() {
		return e2InstanceExecuted;
	}

	public int getE2InstanceWithoutViolation() {
		return e2InstanceWithoutViolation;
	}

	public int getE2InstanceWithViolation() {
		return e2InstanceWithViolation;
	}

	public int getE2AdaptationNotPossible() {
		return e2AdaptationNotPossible;
	}

	public int getE2InstanceWithAdaptation() {
		return e2InstanceWithAdaptation;
	}

	public int getE2FalsePositive() {
		return e2FalsePositive;
	}
	
	public int getE1InstanceExecuted() {
		return e1InstanceExecuted;
	}

	public int getE1InstanceWithoutViolation() {
		return e1InstanceWithoutViolation;
	}

	public int getE1InstanceWithViolation() {
		return e1InstanceWithViolation;
	}

	public int getE1AdaptationNotPossible() {
		return e1AdaptationNotPossible;
	}

	public int getE1InstanceWithAdaptation() {
		return e1InstanceWithAdaptation;
	}

	public int getE1FalsePositive() {
		return e1FalsePositive;
	}

	public int getE3InstanceExecuted() {
		return e3InstanceExecuted;
	}

	public int getE3InstanceWithoutViolation() {
		return e3InstanceWithoutViolation;
	}

	public int getE3InstanceWithViolation() {
		return e3InstanceWithViolation;
	}

	public int getE3AdaptationNotPossible() {
		return e3AdaptationNotPossible;
	}

	public int getE3InstanceWithAdaptation() {
		return e3InstanceWithAdaptation;
	}

	public int getE3FalsePositive() {
		return e3FalsePositive;
	}

	public int getE1PossibleServiceInvocations() {
		return e1PossibleServiceInvocations;
	}

	public int getE1ServiceInvoked() {
		return e1ServiceInvoked;
	}

	public int getE3PossibleServiceInvocations() {
		return e3PossibleServiceInvocations;
	}

	public int getE3ServiceInvoked() {
		return e3ServiceInvoked;
	}

	
	
}